#
# R Function for Multivariate Gaussian Bayes Classifier
#
# This is not a "naive" Bayes classifier, because the
# features are not assumed to be independent. The only
# assumption is a multivariate normal distribution of the features.
#
# Version: 1.1
# Date:    13.11.2013
# License: GPL-2
#
# Author: Christoph Dalitz
#         Institute for Pattern Recognition
#         Niederrhein University of Applied Sciences
#


#
# Arguments:
#   train = train data frame (one feature vector per line)
#   test  = test data frame (one feature vector per line)
#   trainclasses = classes of the train data
#   apriori = when "fromtrain", the a priori probabilities are
#           derived from the trianing data, otherwise they are
#           assumed to be equal for each class
#
# Returns:
#   the list of class names in the order of the test data
#
gaussian.bayes <- function(train, test, trainclasses, apriori="equal") {

  ntrain <- nrow(train)
  ntest <- nrow(test)
  classes <- levels(trainclasses)
  nclasses <- length(classes)

  # train
  meanlist <- list()
  covlist <- list()
  aprioriprob <- rep(1/nclasses, nclasses)
  i <- 0
  for (cl in classes) {
    i <- i + 1
    covlist[[i]] <- cov(train[trainclasses==cl,])
    meanlist[[i]] <- colMeans(train[trainclasses==cl,])
    if (apriori == "fromtrain") {
      aprioriprob[[i]] <- sum(trainclasses==cl) / ntrain
    }
  }
  
  # classify
  result <- rep("", ntest)
  for (j in 1:ntest) {
    t <- test[j,]
    maxprob <- 0
    maxclass <- ""
    for (i in 1:length(classes)) {
      prob <- aprioriprob[[i]] * mvnormdensity(t, meanlist[[i]], covlist[[i]])
      if (prob > maxprob) {
        maxprob <- prob
        maxclass <- classes[[i]]
      }
    }
    result[[j]] <- maxclass
  }

  return(result)
}


#
# density function of the multivariate normal distribution
# Comment: we could use mvtnorm::dmvnorm instead,
#          but package mvtnorm is not installed by default
#
mvnormdensity <- function(x, mean, covar) {
  if (is.vector(x)) {
    x <- matrix(x, ncol = length(x))
  }
  if (ncol(x) != length(mean)) {
    stop("x and mean must have same dimension")
  }
  if (ncol(x) != ncol(covar)) {
    stop("x and covar must have same dimension")
  }
  return( (1 / sqrt((2*pi)^ncol(x) * det(covar))) *
         exp( -mahalanobis(x, center=mean, cov=covar) / 2 ) )
}
