#
# R functions for computing the scatter matrices Sb and Sw from a data set
#
# Version: 1.2
# Date:    09.01.2014
# License: GPL-2
#
# Author: Christoph Dalitz
#         Institute for Pattern Recognition
#         Niederrhein University of Applied Sciences
#

#==========================================================
# Arguments:
#   data = data frame (one feature vector per line)
#   dataclasses = classes of the lines in the data frame
#
# Returns:
#   the between scatter matrix Sb
#
scatter.between <- function(data, dataclasses) {

  classes <- levels(factor(dataclasses))
  nfeatures <- ncol(data)

  Sb <- matrix(rep(0, nfeatures*nfeatures), nrow=nfeatures, ncol=nfeatures)
  m <- colMeans(data)
  for (cl in classes) {
    dc <- data.frame(data[dataclasses==cl,])
    mc <- colMeans(dc)
    Sb <- Sb + nrow(dc)*(mc-m)%o%(mc-m)
  }

  return(Sb)
}

#==========================================================
# Arguments:
#   data = data frame (one feature vector per line)
#   dataclasses = classes of the lines in the data frame
#
# Returns:
#   the within scatter matrix Sw
#
scatter.within <- function(data, dataclasses) {

  classes <- levels(factor(dataclasses))
  nfeatures <- ncol(data)

  Sw <- matrix(rep(0, nfeatures*nfeatures), nrow=nfeatures, ncol=nfeatures)
  for (cl in classes) {
    dc <- as.matrix(data[dataclasses==cl,])
    mc <- colMeans(data.frame(dc))
    for(i in 1:nrow(dc)) {
      x <- dc[i,]
      Sw <- Sw + (x-mc)%o%(x-mc)
    }
  }

  return(Sw)
}

#==========================================================
# Arguments:
#   data = data frame (one feature vector per line)
#   dataclasses = classes of the lines in the data frame
#
# Returns:
#   creates a colored scatter plot
#
scatter.plot <- function(data, dataclasses, title=NA) {
  if (ncol(data) > 2) {
    stop("can only plot 2d data")
  }

  classes <- levels(factor(dataclasses))
  minx = min(data[,1])-0.5
  maxx = max(data[,1])+1.0
  miny = min(data[,2])-0.5
  maxy = max(data[,2])+1.0
  for (i in 1:length(classes)) {
    cl = classes[i]
    if (i == 1) {
      plot(data[dataclasses==cl,1], data[dataclasses==cl,2], xlim=c(minx,maxx), ylim=c(miny,maxy), col=i, main=title)
    } else {
      points(data[dataclasses==cl,1], data[dataclasses==cl,2], col=i)
    }
  }
  legend("topright",classes,lty=c(1,1,1,1),col=1:length(classes))

}

#==========================================================
# Arguments:
#   data = data frame (one feature vector per line)
#   dataclasses = classes of the lines in the data frame
#
# Returns:
#   separation measure J1 = tr(Sb)/tr(Sw)
#
J1 <- function(data, dataclasses) {
  return(sum(diag(scatter.between(data, dataclasses))) /
         sum(diag(scatter.within(data, dataclasses))) )
}

#==========================================================
# Arguments:
#   data = data frame (one feature vector per line)
#   dataclasses = classes of the lines in the data frame
#
# Returns:
#   separation measure J2 = tr(Sw^(-1) * Sb)
#
J2 <- function(data, dataclasses) {
  return(sum(diag(
                  scatter.between(data, dataclasses) *
                  solve(scatter.within(data, dataclasses))
                  )) )
}

# for testing
#if (TRUE) {
if (FALSE) {
  data <- rbind(data.frame(x1=1, x2=0, class="w1"),
                data.frame(x1=2, x2=3, class="w1"),
                data.frame(x1=1, x2=1, class="w1"),
                data.frame(x1=3, x2=0, class="w2"),
                data.frame(x1=4, x2=1, class="w2")
                )
  Sb <- scatter.between(data[,-3],data[,3])
  print("Sb =", quote=FALSE); print(Sb)
  Sw <- scatter.within(data[,-3],data[,3])
  print("Sw =", quote=FALSE); print(Sw)
  print("Sw + Sb =", quote=FALSE); print(Sw+Sb)
  print("(n-1) * cov =", quote=FALSE); print((nrow(data)-1)*cov(data[,-3]))
  scatter.plot(data[,-3],data[,3])
  print(sprintf("J1 = %f", J1(data[,-3],data[,3])), quote=FALSE)
  print(sprintf("J2 = %f", J2(data[,-3],data[,3])), quote=FALSE)
}

